/* -*- mode: c++; c-file-style: "engine"; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/****************************************************************************
 * Copyright Olivier Delannoy, Laurent Choy 2005
 *
 * Olivier Delannoy   
 *   Olivier.Delannoy@gmail.com
 *   PRiSM Laboratory 
 *   Verailles University 
 *   45 avenue des Etats Unis 
 *   78035 Versailles Cedex FRANCE 
 *
 * Laurent Choy 
 *   Choy.Laurent@gmail.com
 *   INRIA Futurs
 *   Parc Club Orsay Universit                      
 *   ZAC des vignes                      
 *   2-4, rue Jacques Monod                      
 *   91893 Orsay Cedex France
 *
 *
 *
 * This software is a computer program whose purpose is to [describe
 * functionalities and technical features of your software].
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 ****************************************************************************/

/**
 * @file
 * @brief Class SchedulerTask declaration
 *  
 * 2005-08-24
 * Olivier DELANNOY
 */
#ifndef YML_CORE_SCHEDULERTASK_HH
#define YML_CORE_SCHEDULERTASK_HH 1
#include "core.hh"
#include <vector>

namespace Yml {
    namespace Core {
        /**
         * @brief This class defines the lifetime of the execution of a task. 
         *
         *
         * This class store all information used during the execution
         * of task. It is created by the scheduler during the
         * execution of the application when the task enter the state
         * ready and is destroyed once the task enter the state
         * finished. by the scheduler too. The mamangement of this
         * object evolves between the Scheduler, the backend manager
         * and the backend. Some field are significant only at certain
         * time. For example, there is no interest in asking for the
         * status of the execution of a task before it's execution. 
         * 
         * @note signal task are transmitted to the BackendManager but
         * not to the backend object. 
         *
         * @note some fields are valid only from a time t. All field
         * with a limited validity have a note. 
         */
        class SchedulerTask
        {
        public:
            /**
             * Default Contstructor
             */
            SchedulerTask(uint32 index);
            
            /**
             * Destructor
             */
            ~SchedulerTask(void);
            /** 
             * Retrieve the index in the task index table 
             * \note can be used in regular back-end 
             */
            uint32 index(void) const;
            /**
             * Retrieve the application id 
             * \note can be used in regular back-end 
             */
            uint32 appId() const;
            uint32& appId();
            /**
             * Retrieve the work description (for the worker) 
             * \note can be used in regular back-end 
             */
            const string& work() const;
            string& work();
            
            /**
             * The number of events 
             */
            uint32 events(void) const;
            /** 
             * Retrieve the event index 
             */
            uint32 event(uint32 index) const;
            /** 
             * Add an event to the list of events generated by the
             * execution of this task.
             */
            void addEvent(uint32 index);
            /** 
             * Check whether the task is a signal or not 
             */
            bool isSignal(void) const;
            /**
             * Check wether the task is a builtin or not 
             */
            bool isBuiltinCall(void) const;
            /**
             * Check wether the task is a call or not
             */
            bool isCall(void) const;
            /**
             * Retrieve the name of the builtin component 
             */
            const string& builtin() const;
            /**
             * Set the name of the builtin component 
             */
            void builtin(const string& name);
            /** 
             * Retrieve the name of the component 
             * \note can be used in regular back-end 
             */
            const string& component(void) const;
            /**
             * Set the name of a component 
             */
            void component(const string& name);
            /** 
             * Retrieve the Scheduling Policy associated with this component 
             * \note can be used in regular back-end 
             */
            SchedulingPolicy policy(void) const;
            /** 
             * Set the policy 
             */
            void policy(SchedulingPolicy policy);
            
            /**
             * Retrieve the number of registered parameters. 
             * \note can be used in regular back-end 
             */
            uint32 parameters(void) const;
            /**
             * Return the @em index parameter description. 
             * \note can be used in regular back-end 
             */
            const string& parameter(uint32 index) const;            
            /** 
             * Add a new parameter description 
             * 
             */ 
            void addParameter(const string& param);
            /**
             * Set the name of all packs 
             */
            void setPackNames(const string& packin, const string& packout, const string& packtmp);
            /*! Set the uid of the task */
            void uid(const string& uid);
            /*!
             * Retrieve the uid of the task 
             * \note can be used in regular back-end 
             */ 
            const string& uid() const;
            /**
             * Retrieve the name of the data input package 
             */
            const string& packInput(void) const;
            /**
             * Retrieve the name of the data output package
             */
            const string& packOutput(void) const;
            /** 
             * Retrieve the name of the admin package
             */
            const string& packAdmin(void) const;
            /** 
             * Retrieve the name of the working directory of the worker (YML-XMP)
             */
            const string& packTmp(void) const;
            /**
             * Retrieve the work description of the task 
             */
            const string& workDescription(void) const;
            /** 
             * Extract administration information from the adminPack
             * \note this method is available once the execution
             * finished. 
             */
            bool extractAdminInfo(void);
            
            /** 
             * Return the status of the execution of a task 
             *
             * \note this information is available once the execution
             * finished. 
             */
            ExecutionStatus status(void) const;
            /** 
             * Allow backends to set the status reported by the backend 
             */
            void setStatus(ExecutionStatus status);

            /**
             * Return the exception name 
             *
             * \note this value is significant only if the
             * status is set to \em ExecException
             */
            const string& exceptionName(void) const;
            /** 
             * Retrieve the trace of the program 
             * 
             * \note use this function only once the execution of the
             * task is finished. 
             */
            const string& trace(void) const;
	    /** 
	     * Add a collection prefix 
             */
	    void addOutputCollection(const string& prefix);
	    /**
	     * Get the number of collection prefix 
	     */
	    uint32 outputCollections() const;
	    /** 
              * Get the collection prefix of \em index
	      */
	    const string& outputCollection(uint32 index) const;
            /* Set and get the number of processes for parallel task */
            void   setNprocs(int n);
            uint32 getNprocs();            
        protected:
        private:
            /** Store the task index */
            uint32 mIndex; //!< Task ID 
            uint32 mAppId; //!< Application ID 
            string mWork;  //!< Work description (input for the worker) 
            /** Store post condition */
            std::vector<uint32> mEvents;
            /** Store the name of the component */
            string mComponent;
            /** Store the name of the builtin component */ 
            string mBuiltin;
            /** Store the scheduling policy */ 
            SchedulingPolicy mPolicy;
            /** Store component parameter descriptions */
            std::vector<string> mParameters;
            /** Store the name of the input package file */
            string mInputPackFile;
            /** Store the name of the output package file */
            string mOutputPackFile;
            /** Store the name of the working directory path of worker */
            string mWorkDir;
            /** Store the name of the admin package file */
            string mAdminPackFile;
            string mWorkDescription;
            /** Store the execution status */
            ExecutionStatus mStatus;
            /** Store the exception name */
            string mExceptionName;
            /** Store the execution trace */
            string mTrace;
            /** Store the uid */
            string mUid;
            /** Store output collection prefixes */
	    std::vector<string> mOutputCollections;
            /** Sotre the number of processes for parallel task **/
            uint32 mNprocs;

            // Disabled fonctions 
            SchedulerTask(SchedulerTask& src);
            SchedulerTask(void);
            SchedulerTask& operator=(SchedulerTask& src);
            
        };
    }
}
#ifdef HAVE_INLINE
#include "SchedulerTask.icc"
#endif
#endif

