/* -*- mode: c++; c-file-style: "engine"; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/****************************************************************************
 * Copyright Olivier Delannoy, Laurent Choy 2005
 *
 * Olivier Delannoy   
 *   Olivier.Delannoy@gmail.com
 *   PRiSM Laboratory 
 *   Verailles University 
 *   45 avenue des Etats Unis 
 *   78035 Versailles Cedex FRANCE 
 *
 * Laurent Choy 
 *   Choy.Laurent@gmail.com
 *   INRIA Futurs
 *   Parc Club Orsay Universit                      
 *   ZAC des vignes                      
 *   2-4, rue Jacques Monod                      
 *   91893 Orsay Cedex France
 *
 *
 *
 * This software is a computer program whose purpose is to [describe
 * functionalities and technical features of your software].
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 ****************************************************************************/

/**
 * @file
 * @brief Class YvetteTaskGenerator function definition 
 *
 * 2005-06-28
 * Olivier Delannoy
 */
#include "YvetteTaskGenerator.hh"
#ifndef HAVE_INLINE
#define inline
#include "YvetteTaskGenerator.icc"
#undef inline
#endif 
// Yml::Core includes 
#include "YvetteCompiler.hh"
#include "AST.hh"
#include "YvetteObject.hh"
#include "YvetteFunction.hh"
#include "YvetteFunctionTable.hh" 
#include "TaskComputation.hh"
#include "PreCondition.hh"
#include "Task.hh" 
#include "YvetteContext.hh"
#include "YvetteParser.hh"

// Util includes 
#include <LoggerFactory.hh>
#include <Application.hh>

// STD includes
#include <algorithm> 
#include <sstream>
#include <iostream> 


Yml::Core::YvetteTaskGenerator::YvetteTaskGenerator(Yml::Core::YvetteCompiler* compiler)
    : mCompiler(compiler)
{
    
}

Yml::Core::YvetteTaskGenerator::~YvetteTaskGenerator(void)
{
    
}


bool Yml::Core::YvetteTaskGenerator::generate(Yml::Core::AST* tree, Yml::Core::YvetteContext& context)
{   
    while(tree)
    {
        bool status = false;
        
        switch(tree->getType())
        {
        case TK_K_ASSIGN:
            status = visitAssignment(tree->getFirstChild());
            break;
            
        case TK_K_COMPUTE:
        case TK_K_MIGRATE:
            status = visitComponentCall(tree->getFirstChild(), tree->getType(), context);
            break;
            
        case TK_K_PAR:
            status = visitPar(tree->getFirstChild(), context);
            break;
            
        case TK_K_SEQ:
            status = visitSeq(tree->getFirstChild(), context);
            break;
            
        case TK_K_IF:
            status = visitConditional(tree->getFirstChild(), context);
            break;
            
        case TK_K_NOTIFY:
            status = visitEventNotify(tree->getFirstChild(), context);
            break;

        case TK_K_WAIT:
            status = visitEventWait(tree->getFirstChild(), context);
            break;
            
        default:
            status = false;
        }   
        if (! status)
        {
            return false;
            
        }
        tree = tree->getNextSibling();
    }
    return true;
}

bool Yml::Core::YvetteTaskGenerator::visitAssignment(Yml::Core::AST* tree)
{
    Yml::Core::string identifier;
    if (visitLValue(tree, identifier))
    {
    
        Yml::Core::YvetteObject value = visitRValue(tree->getNextSibling());
        if (value.isValid())
        {
            //UTIL_OUT() << "Assign '" << value << "' to '" << identifier << "'" << std::endl;
            mCompiler->getSymbolTable().addSymbol(identifier, value);
            return true;
        }
    }
    return false;
}

bool Yml::Core::YvetteTaskGenerator::visitEventWait(Yml::Core::AST* tree, Yml::Core::YvetteContext& context)
{
    //UTIL_OUT() << "visitEventWait(...)" << std::endl;
    PreCondition* condition = visitEventCondition(tree);
    if (condition)
    {
        context.mergePreCondition(condition);
        return true;
        
    }
    return false;
}

bool Yml::Core::YvetteTaskGenerator::visitEventNotify(Yml::Core::AST* tree, Yml::Core::YvetteContext& context)
{
    //UTIL_OUT() << "visitEventNotify(...)" << std::endl;
    // Create a task with the current precondition and postcondition 
    Task *notifyTask = new Task(context);
    while(tree)
    {
        string eventName;
        if (!visitLValue(tree, eventName))
        {
            return false;
        }
        notifyTask->addPostEvent(eventName);
        tree = tree->getNextSibling();        
    }
    string progressEvent = mCompiler->generateProgressEvent();
    
    notifyTask->addPostEvent(progressEvent);
    context.preCondition(new PreCondition(progressEvent));
    mCompiler->getApplication().registerTask(notifyTask);
    return true;
}

bool Yml::Core::YvetteTaskGenerator::visitComponentCall(Yml::Core::AST* tree, Yml::Core::uint32 schedulingPolicy, Yml::Core::YvetteContext& context)
{
    //UTIL_OUT() << "visitComponentCall(...)" << std::endl;
    //Create a task with the current precondition and postcondition 
    //Set the Computation information of the task 
    //  - The scheduling policy 
    //  - The component name
    //  - The parameter list 
    //Register the task. 
    string cName = tree->getText();
    AST *params = tree->getFirstChild();
    TaskComputation* computation = new TaskComputation(schedulingPolicy, cName);
    while(params)
    {
        YvetteObject pvalue = visitRValue(params);
        if (! pvalue.isValid())
        {
            // It's the name of a parameter 
            Yml::Core::string name;
            if (! visitLValue(params, name))
            {
                return false;
            }
            //Add this parameter to the Task Object ;p 
            string prefix("F[] ");
            // Register IT
            computation->addParameter(prefix + name);
        }
        else 
        {
        
            string prefix;
            switch(pvalue.getType())
            {
            case YvetteObject::INTEGER:
                prefix = "C[INTEGER] ";
                break;

            case YvetteObject::REAL:
                prefix = "C[REAL] ";
                break;

            case YvetteObject::STRING:
                prefix = "C[STRING] ";
                break;
            }    
            // Add this parameter as a const value to the Task Object 
            computation->addParameter(prefix + pvalue.stringValue());
        }
        params = params->getNextSibling();
    }
    Task* task = new Task(context);
    task->setComputation(computation);
    // Handle Events
    string progressEvent = mCompiler->generateProgressEvent();
    task->addPostEvent(progressEvent);
    context.preCondition(new PreCondition(progressEvent));
    mCompiler->getApplication().registerTask(task);
    return true;
}

bool Yml::Core::YvetteTaskGenerator::visitConditional(Yml::Core::AST* tree, Yml::Core::YvetteContext& context)
{
    Yml::Core::AST* condNode = tree;
    Yml::Core::AST* thenNode = tree->getNextSibling();
    Yml::Core::AST* elseNode = thenNode->getNextSibling();
    Yml::Core::YvetteObject condValue = visitRValue(condNode);
    if (condValue.isValid())
    {
        if (condValue.isTrue())
        {
            return (thenNode->getFirstChild()) ? generate(thenNode->getFirstChild(), context) : true;
        }
        else 
        {
            return elseNode->getFirstChild() ? generate(elseNode->getFirstChild(), context) : true;
        }
        return true;
    }
    UTIL_FATAL("default", "Invalid Conditional");
    return false;
}

bool Yml::Core::YvetteTaskGenerator::visitPar(Yml::Core::AST* tree, Yml::Core::YvetteContext& context)
{
    //UTIL_OUT() << "visitPar(...)" << std::endl;
    if (tree->getType() == TK_K_DO)
    {
        // Par Do
        Yml::Core::AST* stmt = tree->getFirstChild()->getNextSibling();
        Yml::Core::AST* assign = tree->getFirstChild()->getFirstChild();
        Yml::Core::string symbolName;
        if (!visitLValue(assign->getFirstChild(), symbolName))
        {
            return false;
        }
    

        Yml::Core::YvetteObject limit1= visitRValue(assign->getFirstChild()->getNextSibling());
        Yml::Core::YvetteObject limit2= visitRValue(assign->getNextSibling());
        if (!limit1.isValid() || !limit2.isValid())
        {
            return false;
        }

        Yml::Core::int32 l1 = limit1.integerValue();
        Yml::Core::int32 l2 = limit2.integerValue();
        if (l1 > l2)
        {
            std::swap(l1, l2);
        }
        //UTIL_OUT() << "Iterator[" << l1 << "," << l2 << "]" << std::endl;
        //UTIL_OUT() << context << std::endl;
        YvetteContext resultingContext;
        for(Yml::Core::int32 iter = l1 ; iter <= l2 ; ++iter)
        {
            YvetteContext newContext(context);
            // Copy Initial Pre condition 
            // Copy Initial Post condition
            // Create a new symbol table 
            mCompiler->getSymbolTable().pushContext();
            mCompiler->getSymbolTable().addSymbol(symbolName, Yml::Core::YvetteObject(iter));
            if (! generate(stmt, newContext))
                return false;
            
            // Destroy the symbol table 
            mCompiler->getSymbolTable().popContext();
            
            resultingContext.mergePreCondition(newContext.preCondition());
            //delete newContext.preCondition();
            //Memory Leak here ... To be solved later 
        }
        // Create a new notifyTask 
        Task * notifyTask = new Task(resultingContext);
        string progressEvent = mCompiler->generateProgressEvent();
        notifyTask->addPostEvent(progressEvent);
        context.resetPost();
        context.resetPre();
        
        context.preCondition(new PreCondition(progressEvent));
        mCompiler->getApplication().registerTask(notifyTask);
        //Memory Leak here ... 
    }
    else 
    {
        // Par 
        Yml::Core::AST* blocs = tree;
        YvetteContext resultingContext;
        while(blocs)
        {
            //blocs->getFirstChild()->printTree(UTIL_OUT());
            // Create a copy of the context 
            YvetteContext newContext(context);
            // Create a new symbol table
            mCompiler->getSymbolTable().pushContext();
            generate(blocs->getFirstChild(), newContext);
            // Destroy the symbol table 
            mCompiler->getSymbolTable().popContext();
            // Update the resulting context 
            resultingContext.mergePreCondition(newContext.preCondition()); 
            // Go to the next parallel section 
            blocs = blocs->getNextSibling();
        }
        // Create a new notifyTask 
        Task * notifyTask = new Task(resultingContext);
        string progressEvent = mCompiler->generateProgressEvent();
        notifyTask->addPostEvent(progressEvent);
        context.resetPost();
        context.resetPre();
        delete context.preCondition();
        context.preCondition(new PreCondition(progressEvent));
        mCompiler->getApplication().registerTask(notifyTask);
    }
    return true;
}


bool Yml::Core::YvetteTaskGenerator::visitSeq(Yml::Core::AST* tree, Yml::Core::YvetteContext& context)
{
    //UTIL_OUT() << "visitSeq(...)" << std::endl;
    Yml::Core::AST* stmt = tree->getFirstChild()->getNextSibling();
    Yml::Core::AST* assign = tree->getFirstChild()->getFirstChild();
    Yml::Core::string symbolName;
    if (!visitLValue(assign->getFirstChild(), symbolName))
    {
        return false;
    }
    

    Yml::Core::YvetteObject limit1= visitRValue(assign->getFirstChild()->getNextSibling());
    Yml::Core::YvetteObject limit2= visitRValue(assign->getNextSibling());
    if (!limit1.isValid() || !limit2.isValid())
    {
        return false;
    }
    

    Yml::Core::int32 l1 = limit1.integerValue();
    Yml::Core::int32 l2 = limit2.integerValue();
    if (l1 > l2)
    {
        std::swap(l1, l2);
    }
    //Behave exactly as a default task call and all the information is updated from generate 
    // Nothing to do with the context 
    for(Yml::Core::int32 iter = l1 ; iter <= l2 ; ++iter)
    {
        // Create a new symbol table context 
        mCompiler->getSymbolTable().pushContext();
        // Register the symbolName to the symbol table to the value of Iter 
        mCompiler->getSymbolTable().addSymbol(symbolName, Yml::Core::YvetteObject(iter));
        if (! generate(stmt, context))
        {
            return false;
        }
        // Destroy the symbol table context 
        mCompiler->getSymbolTable().popContext();
    }
    return true;
}

bool Yml::Core::YvetteTaskGenerator::visitLValue(Yml::Core::AST* tree, string& result)
{
    result = tree->getText();
    Yml::Core::AST* arrayControl = tree->getFirstChild();
    while(arrayControl)
    {
        Yml::Core::YvetteObject obj = visitRValue(arrayControl);
        if (obj.isValid() && obj.getType() == Yml::Core::YvetteObject::INTEGER)
            result += "@@" + obj.stringValue();
        else 
            return false;
        arrayControl = arrayControl->getNextSibling();
    }
    //UTIL_OUT() << "visitLValue(...): " << result << std::endl;
    return true;
    
}


Yml::Core::YvetteObject Yml::Core::YvetteTaskGenerator::visitRValue(Yml::Core::AST* tree)
{
    switch(tree->getType())
    {
    case TK_INTEGER:
    {
        std::istringstream stream(tree->getText());   
        Yml::Core::int32 res;
        stream >> res;
        //UTIL_DEBUG("default", "INTEGER<<<" << res << ">>>");
        return Yml::Core::YvetteObject(res);
    }
    case TK_REAL:
    {
        std::istringstream stream(tree->getText());
        Yml::Core::real64 res;
        stream >> res;
        return Yml::Core::YvetteObject(res);
    }
    case TK_STRING:
    {
        return Yml::Core::YvetteObject(tree->getText());
    }
    
    case TK_IDENTIFIER:
    {
        Yml::Core::string varName;
        if (! visitLValue(tree, varName))
        {
            return Yml::Core::YvetteObject::INVALID_OBJECT;   
        }
        return mCompiler->getSymbolTable().getSymbol(varName);
    }
    
    case TK_FUNC:
    {
        //Find the function object 
        Yml::Core::string package = tree->getFirstChild()->getText();
        Yml::Core::string function = tree->getFirstChild()->getFirstChild()->getText();
        Yml::Core::AST* params = tree->getFirstChild()->getNextSibling();
        Yml::Core::YvetteFunctionTable* funcTable = Yml::Core::YvetteFunctionTable::getSingletonPtr();
        // Load the corresponding function package 
        funcTable->loadPackage(package);
        // Retrieve the functor 
        Yml::Core::YvetteFunction* functor = funcTable->get(package, function);
        if (functor != 0)
        {
            for (; params ; params = params->getNextSibling())
            {
            //    Yml::Core::YvetteObject tmp = visitRValue(params);
          //      UTIL_DEBUG("default", "Register param: " << tmp.stringValue());
                functor->addParameter(visitRValue(params));
            }
            return functor->execute();
        }
    }
    }
    
    return Yml::Core::YvetteObject::INVALID_OBJECT; // Oups invalid RValue 
}

Yml::Core::PreCondition* Yml::Core::YvetteTaskGenerator::visitEventCondition(Yml::Core::AST* tree)
{
    if (tree->getType() == TK_IDENTIFIER)
    {
        string eventName;
        if (visitLValue(tree, eventName))
            return new PreCondition(eventName);
    } 
    else 
    {
        string op = tree->getFirstChild()->getFirstChild()->getText();
        if (op.compare("operator_and") == 0)
        {
            AST* left = tree->getFirstChild()->getNextSibling();
            AST* right = left->getNextSibling();
            if (right == 0 || right->getNextSibling())
                return 0;
    
            // Create the result of visitEventCondition 
            PreCondition* pLeft =  visitEventCondition(left);
            PreCondition* pRight = visitEventCondition(right);
            if (pLeft != 0 && pRight != 0)
                return new PreCondition(PreCondition::OP_AND, pLeft, pRight);
        }
        else if (op.compare("operator_or") == 0)
        {
            AST* left = tree->getFirstChild()->getNextSibling();
            AST* right = left->getNextSibling();
            if (right == 0 || right->getNextSibling())
                return 0;
            // Create the result of visitEventCondition 
            PreCondition* pLeft =  visitEventCondition(left);
            PreCondition* pRight = visitEventCondition(right);
            if (pLeft != 0 && pRight != 0)
                return new PreCondition(PreCondition::OP_OR, pLeft, pRight);
        }
        else if (op.compare("operator_xor") == 0)
        {
            AST* left = tree->getFirstChild()->getNextSibling();
            AST* right = left->getNextSibling();
            if (right == 0 || right->getNextSibling())
                return 0;
    
            // Create the result of visitEventCondition 
            PreCondition* pLeft =  visitEventCondition(left);
            PreCondition* pRight = visitEventCondition(right);
            if (pLeft != 0 && pRight != 0)
                return new PreCondition(PreCondition::OP_XOR, pLeft, pRight);
        }
        else if (op.compare("operator_not") == 0)
        {
            AST* left = tree->getFirstChild()->getNextSibling();
            if (left->getNextSibling())
                return 0;
            // Create the result of visitEventCondition 
            PreCondition* negateMe = visitEventCondition(left);
            if (negateMe != 0)
                return new PreCondition(negateMe);
        }
    }
    return 0;    
}

